'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.filter = filter;
exports.getLogs = getLogs;

var _bluebird = require('bluebird');

var _bluebird2 = _interopRequireDefault(_bluebird);

var _config = require('../config');

var _config2 = _interopRequireDefault(_config);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * @function filter
 *
 * @description Gets filter object, that can be used for getting of past logs or watching of new
 *
 * @param {object} [options={}] - Filter options
 * @param {(number|string)} [options.fromBlock] - The number of the earliest block
 * @param {(number|string)} [options.toBlock] - The number of the latest block
 * @param {(string|string[])} [options.address] - An address(es) to get logs from
 * @param {string[]} [options.topics] - Allows to manually set the topics for the event filter
 *
 * @returns {object} (@see {@link https://github.com/ethereum/wiki/wiki/JavaScript-API#web3ethfilter})
 */
/**
 * @file Manages helper functions to work with logs
 * @author Ivan Violentov <ivan.violentov@jibrel.network>
 */

function filter() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  return jWeb3.eth.filter(options);
}

/**
 * @async
 * @function getLogs
 *
 * @description Gets past logs
 *
 * @param {object} [options={}] - Filter options
 * @param {(number|string)} [options.fromBlock] - The number of the earliest block
 * @param {(number|string)} [options.toBlock] - The number of the latest block
 * @param {(string|string[])} [options.address] - An address(es) to get logs from
 * @param {string[]} [options.topics] - Allows to manually set the topics for the event filter
 *
 * @returns Promise that will be resolved with past logs
 */
function getLogs() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var filterObject = jWeb3.eth.filter(options);

  /**
   * filter.get uses instance methods inside,
   * but bluebird promisify don't save context,
   * so need to bind to filter object directly
   */
  return _bluebird2.default.promisify(filterObject.get.bind(filterObject))().timeout(_config2.default.promiseTimeout, new Error('Can not get past logs within ' + _config2.default.promiseTimeout + 'ms'));
}